/*
 * Functions supporting channels.
 *
 * Part of Chatserv, by Jonathan Lennox.  Written for no particularly good
 * reason, because I felt like it.  Allows multiple people to chat over the
 * net in a relatively straightforward environment.  Disclaimer: I figured out
 * how to do TCP/IP programming by reading man pages.
 *
 * Copyright 1995 Jonathan Lennox
 *
 * See LICENSE for licensing terms.
 */

#ifdef HAVE_CONFIG_H
# include <config.h>
#endif

#if STDC_HEADERS
# include <string.h>
# include <stdlib.h>
#else
# if HAVE_STRING_H
#  include <string.h>
# else
#  include <strings.h>
# endif
# if HAVE_STDLIB_H
#  include <stdlib.h>
# else
#  include <malloc.h>
# endif
#endif
#if TIME_WITH_SYS_TIME
# include <sys/time.h>
# include <time.h>
#else
# if HAVE_SYS_TIME_H
#  include <sys/time.h>
# else
#  include <time.h>
# endif
#endif

#include <stdio.h>
#include <errno.h>
#include <ctype.h>
#include "sunos-protos.h"

#include "channels.h"
#include "connections.h"
#include "nicks.h"
#include "config-file.h"
#include "format.h"
#include "dyn-str.h"

/* Global channel-related variables */
struct channel_heads channels[NUM_CHANNELS];
int default_channel;

/* Some of this stuff is kinda boring now, but I wanna provide an
   abstract API in case the "channel" struct ever changes.
*/

/* Set up the global channels.  Returns -ERRNO on errors. */
int set_up_channels() {
  int i;
  char lookup_val[17]; /* Assumes <= 99 channels */
  char *val;

  for (i = 0; i < NUM_CHANNELS; i++) {
    channels[i].head = NULL;
    sprintf(lookup_val, "channel_%d_topic", i);
    val = lookup_conf_with_fallback(lookup_val, "default_channel_topic");
    if (val != NULL) {
      channels[i].desc = strdup(val);
      if (channels[i].desc == NULL) return -ENOMEM;
    } else {
      channels[i].desc = NULL;
    }
    channels[i].connection_count = 0;
    channels[i].last_nonidle = time(NULL);
  }
  return 0;
}

int init_connection_channels(struct connection *newcon)
{
  int i;

  newcon->channels.primary_channel = UNCHANNEL;
  for (i = 0; i < NUM_CHANNELS; i++) {
    newcon->channels.next[i] = NULL;
    newcon->channels.prev[i] = NULL;
  }
  newcon->channels.channel_map = 0;

  return 0;
}

/* The number of channels that currently exist. */
int num_channels()
{
  return NUM_CHANNELS;
}

/* What channel is the user currently talking to? */
int primary_channel(struct connection *user)
{
  return user->channels.primary_channel;
}

/* Is this a valid channel? */
/* XXX should the Unchannel be considered a "valid" channel? */
int valid_channel(int channel)
{
  return (channel >= 0 && channel < NUM_CHANNELS);
}

/* Is the user on only one channel right now? */
static int on_only_one_channel(struct connection *user)
{
  int i, count = 0;
  for (i = 0; i < NUM_CHANNELS; i++) {
    if (user->channels.channel_map & (1 << i)) {
      count++;
      if (count > 1) return 0;
    }
  }
  return (count == 1);
}

/* Join (listening to) a channel.  Currently adds the user at the
   start first in the list.  Non-zero on errors.*/
int join(struct connection *user, int channel)
{
  if (!valid_channel(channel)) return -EINVAL;
  if (user->channels.next[channel] != NULL ||
      user->channels.prev[channel] != NULL ||
      (user->channels.channel_map & (1 << channel))) {
    /* Shouldn't happen */
    fprintf(stderr, "User %s tried to join channel %d -- seems to "
	    "already be on it?\n", nickname(user), channel);
  }
  user->channels.next[channel] = channels[channel].head;
  user->channels.prev[channel] = NULL;
  if (channels[channel].head)
    channels[channel].head->channels.prev[channel] = user;
  channels[channel].head = user;
  user->channels.channel_map |= (1 << channel);
  channels[channel].connection_count++;
  return 0;
}

/* Leave (listening to) a channel.  Non-zero on errors. -errno on real
   errors, positive on confusion. */
int leave(struct connection *user, int channel)
{
  if (!valid_channel(channel)) return -EINVAL;
  if (!(user->channels.channel_map & (1 << channel))) {
    /* Shouldn't happen */
    fprintf(stderr, "Uh-oh...user %s trying to leave channel %d, "
	    "but doesn't seem to be on it in the first place!\n",
	    nickname(user), channel);
    return 1;
  }

  if (channels[channel].head == user)
    channels[channel].head = user->channels.next[channel];
  if (user->channels.next[channel])
    user->channels.next[channel]->channels.prev[channel]= 
      user->channels.prev[channel];
  if (user->channels.prev[channel])
    user->channels.prev[channel]->channels.next[channel] = 
      user->channels.next[channel];
  user->channels.next[channel] = NULL;
  user->channels.prev[channel] = NULL;
  user->channels.channel_map &= ~(1 << channel);
  channels[channel].connection_count--;
  if(!channels[channel].connection_count) {
    char *res, s[64];
    sprintf(s, "channel_%d_snapback", channel);
    res = lookup_conf_with_fallback(s, "default_channel_snapback");
    if(*res == 'y' || *res == 'Y') {
      sprintf(s, "channel_%d_topic", channel);
      res = lookup_conf_with_fallback(s, "default_channel_topic");
      if(res) set_channel_topic(res, channel);
    }
  }
  return 0;
}

/* Higher level channel-changing function. */
int change_channels(struct connection *user, int new_primary,
		    int new_map)
{
  int retval, i;
  int saved_channel;
  char chan_str[3]; /* Sufficient for representing any channel number */

  for (i = 0; i < NUM_CHANNELS; i++)
    if ((user->channels.channel_map & (1 << i)) && !(new_map & (1 << i))) {
      /* Currently on the channel, we don't want to be, so leave the
	 channel. */

      retval = visible_action(user);
      if (retval) return retval;

      /* Set the primary_channel value so format can grab from it. */
      saved_channel = user->channels.primary_channel;
      user->channels.primary_channel = i;
      retval = write_to_channel(formatn("leave_channel",
				       user, NULL),
				formatn("leave_channel",
				       user, NULL), 
				i);
      user->channels.primary_channel = saved_channel;
      if (retval) return retval;
      
      retval = leave(user, i);
      if (retval) return retval;
    }
  for (i = 0; i < NUM_CHANNELS; i++)
    if (!(user->channels.channel_map & (1 << i)) && (new_map & (1 << i))) {
      /* Not currently on the channel, we want to be, so join it. */

      retval = visible_action(user);
      if (retval) return retval;

      sprintf(chan_str, "%d", i);
#if 0     /* XXX do I want this? */
      retval = list_cmd("LIST", user, chan_str);
      if (retval) return retval;
#endif

      retval = join(user, i);
      if (retval) return retval;

      /* Set the primary_channel value so format can grab from it. */
      saved_channel = user->channels.primary_channel;
      user->channels.primary_channel = i;
      retval = write_to_channel(formatn("join_channel",
				       user, NULL),
				formatn("join_channel",
				       user, NULL),
				i);
      user->channels.primary_channel = saved_channel;
      if (retval) return retval;

    }

  if ((valid_channel(new_primary) || new_primary == -1) &&
      (user->channels.primary_channel != new_primary)) {
    user->channels.primary_channel = new_primary;
    if (new_primary != -1)
      return write_to_user(formatn("new_primary_channel",
				  user, NULL),
			   user);
    else
      return write_to_user(formatn("new_no_primary_channel",
				  user, NULL),
			   user);
  }
  
  return 0;
}    
    
/* Write a string to all users on a given channel.  -ERRNO on errors. */
int write_to_channel(char *on_chan, char *off_chan, int channel)
{
  struct connection *current;
  int error, old_error = 0;

  if (!valid_channel(channel)) return -EINVAL;
  for (current = channels[channel].head; current != NULL;
       current = current->channels.next[channel])
    {
      if (current->print_blank) {
	error = write_to_user(blankline, current);
	if (error) {
	  old_error = error;
	  continue;
	}
      }
      if ((current->channels.primary_channel == channel) &&
          !(current->client_flags & CLIENT_NODEFAULT))
	error = write_to_user(on_chan, current);
      else
	error = write_to_user(off_chan, current);
      if (error) old_error = error;
    }
  return old_error;
}

/* Write a string to all users on a channel matching a given channelmap.
   -ERRNO on error. */
int write_to_channelmap(char *string, int map)
{
  struct connection *p;
  int retval, old_retval = 0;

  for (p = connections; p != NULL; p = p->next) {
    if (p->channels.channel_map & map) {
      if (p->print_blank) {
	retval = write_to_user(blankline, p);
	if (retval) {
	  old_retval = retval;
	  continue;
	}
      }
      retval = write_to_user(string, p);
      if (retval) old_retval = retval;
    }
  }
  return old_retval;
}


int close_all_channels(struct connection *user)
{
  int i, retval = 0, newretval;
  /* XXX Might want to use higher-level functions? */
  
  for (i = 0; i < NUM_CHANNELS; i++) {
    if (user->channels.channel_map & (1 << i)) {
      newretval = leave(user, i);
      if (newretval) retval = newretval;
    }
  }

  user->channels.primary_channel = -1;

  return retval;
}

/* Get the topic of the specified channel. NULL on error or missing. */
char *get_channel_topic(int channel)
{
  if (!valid_channel(channel)) return NULL;

  return channels[channel].desc;
}

/* Set the topic of the specified channel. -errno on error. */
int set_channel_topic(char *topic, int channel)
{
  if (!valid_channel(channel)) return -EINVAL;

  if (channels[channel].desc != NULL)
    free(channels[channel].desc);
  channels[channel].desc = strdup(topic);
  if (channels[channel].desc == NULL)
    return -ENOMEM;
  return 0;
}

/* Reset the topics of all the channels to their config'd values.
   Typically called after reloading the configuration file.
*/
int reset_channel_topics(void)
{
  int i, retval = 0;
  char lookup_val[17]; /* Assumes <= 99 channels */
  char *val;

  for (i = 0; i < NUM_CHANNELS; i++) {
    if (channels[i].desc != NULL)
      free(channels[i].desc);

    sprintf(lookup_val, "channel_%d_topic", i);
    val = lookup_conf_with_fallback(lookup_val, "default_channel_topic");
    if (val != NULL) {
      channels[i].desc = strdup(val);
      if (channels[i].desc == NULL) retval = -ENOMEM;
    } else
      channels[i].desc = NULL;
  }
  return retval;
}

/* There is activity on a channel. */
int channel_nonidle(int channel)
{
  if (!valid_channel(channel)) return -EINVAL;

  channels[channel].last_nonidle = time(NULL);
  return 0;
}

/* Interpret a set of arguments as a channel, nick, or yell, and give
   the rest of the arguments.  Plops a null down in the string, so be
   warned. */
struct recipient identify_recipient(int allowed, struct connection *user,
				    char *args, char **rest)
{
  char *ptr, *p;
  int ch;
  struct recipient return_me;
  struct connection *rcpt;

  return_me.explicit = 1;

  for (ptr = args; *ptr != '\0' && !isspace(*ptr); ptr++)
    ; /* Do nothing */

  if (*ptr != '\0') {
    *ptr = '\0';
    ptr++;
  }

  if (ptr == args) {
    /* Nothing there. */
    if (rest != NULL)
      *rest = ptr;
    if ((allowed & RCPT_UNSPEC) && (allowed & (user->target.argtype)))
      return user->target;
    else {
      return_me.argtype = RCPT_UNSPEC;
      return return_me;
    }
  }

  for (; *ptr != '\0' && isspace(*ptr); ptr++)
    ; /* Do nothing */

  if (rest != NULL)
    /* "rest" points at the rest of the line, if we weren't passed a NULL. */
    *rest = ptr;
  
  if (allowed & RCPT_CHANNEL && isdigit(*args)) {
    /* This looks like a channel. */
    ch = strtol(args, &p, 10);

    if (args != p && *p == '\0' && valid_channel(ch)) {
      return_me.argtype = RCPT_CHANNEL;
      return_me.u.channel = ch;
      return return_me;
    }
  } 
  if (allowed & RCPT_EVERYONE && args[0] == '*' && args[1] == '\0') {
    /* This looks like everyone. */
    return_me.argtype = RCPT_EVERYONE;
    return return_me;
  } 
  if (allowed & RCPT_USER) {
    /* This looks like a user. */
    rcpt = lookup_nickname(args);
    if (rcpt == NULL ||
	(rcpt->status != normal && user->privs <= normal_priv 
	 && rcpt->status != hiding && rcpt != user)) {
      return_me.argtype = 0;
    } else {
      return_me.argtype = RCPT_USER;
      return_me.u.user = rcpt;
    }
    return return_me;
  }
  return_me.argtype = 0;
  return return_me;
}

/* Commands */

/* Change channel, or list channels */
int channel_cmd(char *cmd_name, struct connection *user, char *args)
{
  char *ptr, *optr;
  int ch, new_ch, new_chmap, retval, i;
  struct connection placeholder;

  if (strcmp(args, "") == 0) {
    if (user->channels.channel_map == 0) {
      retval = write_to_user(formatn("no_current_channels",
				    user, NULL),
			     user);
      if (retval) return retval;
    } else {
      retval = write_to_user(formatn("current_channels",
				    user, NULL),
			     user);
      if (retval) return retval;
    }
    if (valid_channel(user->channels.primary_channel)) {
      retval = write_to_user(formatn("primary_channel",
				    user, NULL),
			     user);
      if (retval) return retval;
    } else {
      retval = write_to_user(formatn("no_primary_channel",
				    user, NULL),
			     user);
      if (retval) return retval;
    }
    

    retval = write_to_user(formatn("channel_list_header",
				  user, NULL),
			   user);
    if (retval) return retval;
    for (i = 0; i < NUM_CHANNELS; i++) {
      placeholder.channels.primary_channel = i;
      retval = write_to_user(formatn("channel_list_entry",
				    &placeholder, NULL),
			     user);
      if (retval) return retval;
    }
    return 0;
  }

  new_chmap = 0; /* user->channels.channel_map; */

  optr = args;
  /* Argument 1.  Indicates the new primary channel */
  ch = strtol(args, &ptr, 0);
  if (optr != ptr) {
    if (valid_channel(ch)) {
      new_chmap |= (1 << ch);
      new_ch = ch;
    } else
      new_ch = -1;
    optr = ptr;
    /* Argument 2.  If it exists, we do a join-channels rather than
       change-channel. */
    ch = strtol(optr, &ptr, 0);
    if (optr != ptr) {
      if (valid_channel(ch))
	new_chmap |= (1 << ch);
      new_chmap |= user->channels.channel_map;
      do {
	optr = ptr;
	ch = strtol(optr, &ptr, 0);
	if (optr != ptr && valid_channel(ch))
	  new_chmap |= (1 << ch);
      } while (optr != ptr);
    } else
      /* We didn't have an argument 2, but if we were on more than one
	 channel before we also do a join rather than a change. */
      if (!on_only_one_channel(user))
	new_chmap |= user->channels.channel_map;
  } else
    new_ch = -1;
  return change_channels(user, new_ch, new_chmap);
}

/* Join a set of channels */
int join_cmd(char *cmd_name, struct connection *user, char *args)
{
  char *ptr, *optr;
  int ch, new_ch, new_chmap;

  if (args[0] == '\0')
    return 0;

  new_chmap = user->channels.channel_map;

  optr = args;
  ch = strtol(args, &ptr, 0);
  if (optr != ptr) {
    if (valid_channel(ch)) {
      new_chmap |= (1 << ch);
      new_ch = ch;
    } else
      new_ch = -1;
    do {
      optr = ptr;
      ch = strtol(optr, &ptr, 0);
      if (optr != ptr && valid_channel(ch))
	new_chmap |= (1 << ch);
    } while (optr != ptr);
  } else
    new_ch = -1;
  return change_channels(user, new_ch, new_chmap);
}

/* Leave a set of channels */
int leave_cmd(char *cmd_name, struct connection *user, char *args)
{
  char *ptr, *optr;
  int ch, new_ch, new_chmap, i;

  new_chmap = user->channels.channel_map;
  
  ptr = args;
  optr = NULL;
  while (ptr != optr) {
    optr = ptr;
    ch = strtol(optr, &ptr, 0);
    if (optr != ptr && valid_channel(ch))
      new_chmap &= ~(1 << ch);
  }
  if (user->channels.primary_channel != -1 &&
      !(new_chmap & (1 << user->channels.primary_channel))) {
    new_ch = -1;
    for (i = 0; i < num_channels(); i++) {
      if (new_chmap & (1 << i)) {
	new_ch = i;
	break;
      }
    }
  } else {
    new_ch = user->channels.primary_channel;
  }
  return change_channels(user, new_ch, new_chmap);
}

/* Leave all your channels and vanish from the listall list. */
int hide_cmd(char *cmd_name, struct connection *user, char *args)
{
  int retval, was_already = 0;

  retval = change_channels(user, -1, 0);
  if(retval) return retval;
  if(user->status == hiding) was_already = 1; else user->status = hiding;
  retval = write_to_user(formatvn(was_already ? "already_hidden" :
                                  "now_hidden", user, 0), user);
  return retval;
}

#ifdef DO_CLOAK
int cloak_cmd(char *cmd_name, struct connection *user, char *args)
{
  int retval;

  retval = write_to_user(formatvn(user->status == cloaked ? "already_cloaked" :
                                  "now_cloaked", user, 0), user);
  user->status = cloaked;
  return retval;
}

int uncloak_cmd(char *cmd_name, struct connection *user, char *args)
{
  int retval;

  user->status = hiding;
  return(visible_action(user));
  return retval;
}
#endif

static int print_target_info(struct connection *user, char *args, int argtype)
{
  int retval;
  struct recipient target;
  struct connection *current, placeholder;

  target = identify_recipient(RCPT_USER | RCPT_CHANNEL | RCPT_EVERYONE |
			      RCPT_UNSPEC,
			      user, args, NULL);

  if (target.explicit == 0 || target.argtype == RCPT_UNSPEC)
    /* We've had no argument explicitly specified, default based on how we
       were called. */
    if (argtype != target.argtype)
      switch (argtype) {
      case RCPT_USER:
	target.argtype = RCPT_USER;
	target.u.user = user;
	break;
      case RCPT_CHANNEL:
	if (valid_channel(user->channels.primary_channel)) {
	  target.argtype = RCPT_CHANNEL;
	  target.u.channel = user->channels.primary_channel;
	}
	break;
      case RCPT_EVERYONE:
	target.argtype = RCPT_EVERYONE;
	break;
      default:
	/* Shouldn't happen */
	target.argtype = 0;
	break;
      }
  
  switch(target.argtype) {
  case RCPT_USER:
    /* This does a "whois" */
    if (target.u.user->hostname_resolved <= 0)
      retval = write_to_user(formatn("whois_origin_no_dns",
				    target.u.user, NULL),
			     user);
    else
      retval = write_to_user(formatn("whois_origin_dns",
				    target.u.user, NULL),
			     user);
    if (retval) return retval;
    
    if (target.u.user->channels.channel_map == 0)
      retval = write_to_user(formatn("whois_channels_none",
				    target.u.user, NULL),
			     user);
    else if (!valid_channel(target.u.user->channels.primary_channel))
      retval = write_to_user(formatn("whois_channels_no_primary",
				    target.u.user, NULL),
			     user);
    else
      retval = write_to_user(formatn("whois_channels_full",
				    target.u.user, NULL),
			     user);
    if (retval) return retval;
    
    retval = write_to_user(formatn("whois_time",
				  target.u.user, NULL),
			   user);
    if (retval) return retval;
    
    return 0;
    break;

  case RCPT_CHANNEL:
    /* This does a "list" */
    placeholder.channels.primary_channel = target.u.channel;
    retval = write_to_user(formatn("list_header",
				  &placeholder, NULL),
			   user);
    if (retval) return retval;

    for (current = channels[target.u.channel].head;
	 current != NULL;
	 current = current->channels.next[target.u.channel]) {
      retval = write_to_user(formatn("list_format", current, NULL),
			     user);
      if (retval) return retval;
    }
    return 0;
    break;

  case RCPT_EVERYONE:
    retval = write_to_user(formatn("listall_header", user, NULL),
			   user);
    if (retval) return retval;

    for (current = connections; current != NULL;
	 current = current->next) {
      if (current->status == normal || user->privs > normal_priv) {
	/* Don't show people in weird states, except to priv'd people */
	retval = write_to_user(formatn("list_format",
				      current, NULL),
			       user);
	if (retval) return retval;
      }
    }
    return 0;
    break;

  case RCPT_UNSPEC:
    /* This should only happen in the case where we have no primary
       channel. */
    /* XXX: Am I sure?  This error message could be somewhat odd otherwise. */
    return write_to_user(formatn("no_primary_channel",
				user, NULL),
			 user);
    break;

  default:
    switch (argtype) {
    case RCPT_USER:
      return write_to_user(formatn("unknown_user", user, args),
			   user);
      break;
      
    case RCPT_CHANNEL:
    case RCPT_EVERYONE:
      return write_to_user(formatn("invalid_channel", user, args),
			   user);
      break;

    default: ;
      /* This shouldn't happen. */
    }
  }
  return 0;
}

/* List users on your current channel */
int list_cmd(char *cmd_name, struct connection *user, char *args)
{
  return print_target_info(user, args, RCPT_CHANNEL);
}
    
/* List all users on the server */
int listall_cmd(char *cmd_name, struct connection *user, char *args)
{
  return print_target_info(user, args, RCPT_EVERYONE);
}

/* Get information on another user, or on yourself. */
int whois_cmd(char *cmd_name, struct connection *user, char *args)
{
  return print_target_info(user, args, RCPT_USER);
}

static dyn_string *list_users(dyn_string *buf, int chan)
{
  struct connection *current;

  buf = dyn_strcpy(buf, "");
  if (buf == NULL) return NULL;

  for (current = channels[chan].head;
       current != NULL;
       current = current->channels.next[chan]) {
    
    if (current != channels[chan].head) {
      buf = dyn_strcat(buf, " ");
      if (buf == NULL) return NULL;
    }
    buf = dyn_strcat(buf, nickname(current));
    if (buf == NULL) return NULL;
  }
  
  return buf;
}

/* Show short-form list of people on a channel or on the server. */
int users_cmd(char *cmd_name, struct connection *user, char *args)
{
  static dyn_string buffer = DYN_STR_INIT(64), *p;
  int chan;
  int retval;
  struct connection placeholder;
  struct recipient target;

  target = identify_recipient(RCPT_CHANNEL | RCPT_EVERYONE | RCPT_UNSPEC, 
			      user, args, NULL);

  if (target.argtype == RCPT_EVERYONE) {
    for (chan = 0; chan < NUM_CHANNELS; chan++) {
      if (channels[chan].connection_count > 0) {
	placeholder.channels.primary_channel = chan;
	p = list_users(&buffer, chan);
	if (p == NULL) return -ENOMEM;
	retval = write_to_user(formatn("users_format",
				      &placeholder, buffer.buf),
			       user);
	if (retval) return retval;
      }
    }
    return 0;
  } else if (target.argtype == RCPT_CHANNEL) {
    placeholder.channels.primary_channel = target.u.channel;
    p = list_users(&buffer, target.u.channel);
    if (p == NULL) return -ENOMEM;
    return write_to_user(formatn("users_format",
				&placeholder, buffer.buf),
			 user);
  } else if (target.argtype == RCPT_UNSPEC) {
    return write_to_user(formatn("no_primary_channel", user, NULL),
			 user);
  } else if (target.argtype == RCPT_USER) {
    return write_to_user(formatn("bad_recipient", user, cmd_name),
			 user);
  } else {
    return write_to_user(formatn("invalid_channel", user, args),
			 user);
  }
}

/* Display or change the current channel topic */
int topic_cmd(char *cmd_name, struct connection *user, char *args)
{
  int retval = 0;
  int chan;

  if (user->target.argtype == RCPT_UNSPEC) {
    return write_to_user(formatn("no_primary_channel", user, NULL),
			 user);
  }

  if (user->target.argtype != RCPT_CHANNEL) {
    return write_to_user(formatn("bad_recipient", user, cmd_name),
			 user);
  }

  chan = user->channels.primary_channel;
  user->channels.primary_channel = user->target.u.channel;

  if (strcmp(args, "") == 0) {
    retval = write_to_user(formatn("channel_topic", user, NULL),
			   user);
    user->channels.primary_channel = chan;
    return retval;
  }

  /* XXX should privs be required? */
  /* XXX should it be possible to set the topic to the empty string? */
  if ((user->channels.channel_map) & (1 << (user->target.u.channel))) {
    retval = set_channel_topic(args, user->channels.primary_channel);
    if (!retval) {
      retval = visible_action(user);
      if (!retval) {
	retval = channel_nonidle(user->channels.primary_channel);
	if (!retval) {
	  retval =
	    write_to_channel(formatn("channel_topic_changed",
				    user, NULL), 
			     formatn("channel_topic_changed",
				    user, NULL), 
			     user->channels.primary_channel);
	}
      }
    }
  } else {
    retval = write_to_user(formatn("couldnt_change_topic",
				  user, NULL),
			   user);
  }
  user->channels.primary_channel = chan;
  return retval;
}
