/*
 * Header files for user connections.
 *
 * Include channels.h first.
 *
 * Part of Chatserv, by Jonathan Lennox.  Written for no particularly good
 * reason, because I felt like it.  Allows multiple people to chat over the
 * net in a relatively straightforward environment.  Disclaimer: I figured out
 * how to do TCP/IP programming by reading man pages.
 *
 * Copyright 1995 Jonathan Lennox
 *
 * See LICENSE for licensing terms.
 */

#include <sys/types.h>
#include <netinet/in.h>

#include "ssl.h"

/* The basic buffer for connections. */
#define BUFFERSIZE 2048

/* The output buffer for connections. */
#define OUTPUT_BUFSIZE 10240

/* The maximum length a message can be. */
#define MAXMSGLENGTH 1024

enum priv { no_priv = -2,
	      logging_on_priv = -1,
	      normal_priv,
	      gm_priv,
	      op_priv,
              srv_priv,
	      LAST_PRIV
	  } privs;

#define MAX_NICK_LENGTH 64

/* A basic connection */
struct connection {
  int fd;
  struct sockaddr_in connection;
  char *hostname;
  int hostname_resolved; /* 0 = in progress; 1 = "hostname" field good;
			    -1 = failed. */
  int reap_me;
  struct channels channels;
  int columns;
  int print_blank;
  int client_flags;
#ifdef DO_UNORTHODOX
  char **unorthoname;
  int unorthocount;
  int unorthodox;
#endif
  char *nick;
  char buffer[BUFFERSIZE];
  int buf_end;
  int has_overflowed;
  enum priv privs;
  enum {unnamed, 
        unrevealed, 
        normal, 
        hiding, 
        cloaked, 
        gone_too_soon, 
        gone} status;
  time_t signon;
  time_t last_nonidle;
  struct connection *prev;
  struct connection *next;
  struct recipient target;
#ifdef DO_ALIASES
  char alias[MAX_NICK_LENGTH + 1];
#endif
  SSL ssl;
  int ssl_finished;
  char *error;

  char output_buffer[OUTPUT_BUFSIZE];
  int waiting;
};

/* For client_flags */

#define CLIENT_NUMERICS 0x00000001
#define CLIENT_NODEFAULT 0x00000002
#define CLIENT_NOSPAM 0x00000004

/* The head of the linked list of extant connections */
extern struct connection * connections;

/* Whether we need to look for connections to reap */
extern int reap;

/* Just "\n", but write_wrapped requires that its argument be writable. */
extern char blankline[];

/* Exported functions defined in connections.c */
int tcp_listen_on(int port);
int set_up_regexps(void);
struct connection * setup_new_connection(int fd);
#ifdef DO_SSL
struct connection * setup_new_ssl_connection(int fd);
#endif
int close_connection(struct connection *user);
int greet_new_user(struct connection *user);
int read_from_user(struct connection *user);
int parse_user_line(char *line, struct connection *user);
int send_to_recipient(struct connection *user, char *line, char *msg_type,
		      struct recipient rcpt);
int write_to_user(char *string, struct connection *user);
int announce_connection(struct connection *user);
int announce_disconnection(struct connection *user, char *reason);
int visible_action(struct connection *user);
int write_to_everyone(char *string);
#ifdef DO_SPAM
int write_to_everyone_checkspam(char *string, int map);
#else
#define write_to_everyone_checkspam(s, m) write_to_everyone(s)
#endif
void attempt_write(struct connection *user);

/* Telnet needs this. */
int queue_for_write(struct connection *user, char *string, int len);

int send_to_other_channel(struct connection *user, int chan, char *args);
int pose_to_other_channel(struct connection *user, int chan, char *args);

void handler(int sig);

int report_client_numeric_info(struct connection *user);
int report_client_formats();

int bye_cmd(char *cmd_name, struct connection *user, char *args);
int pose_cmd(char *cmd_name, struct connection *user, char *args);
int ppose_cmd(char *cmd_name, struct connection *user, char *args);
int yay_cmd(char *cmd_name, struct connection *user, char *args);
int yell_cmd(char *cmd_name, struct connection *user, char *args);
int as_cmd(char *cmd_name, struct connection *user, char *args);
int narrate_cmd(char *cmd_name, struct connection *user, char *args);
int whisper_cmd(char *cmd_name, struct connection *user, char *args);
int wrap_cmd(char *cmd_name, struct connection *user, char *args);
int client_cmd(char *cmd_name, struct connection *user, char *args);
int kick_cmd(char *cmd_name, struct connection *user, char *args);

int start_dns();
void do_dns_received(int fd);
