/*
 * Handle pretty formatted text.
 *
 * Part of Chatserv, by Jonathan Lennox.  Written for no particularly good
 * reason, because I felt like it.  Allows multiple people to chat over the
 * net in a relatively straightforward environment.  Disclaimer: I figured out
 * how to do TCP/IP programming by reading man pages.
 *
 * Copyright 1995 Jonathan Lennox
 *
 * See LICENSE for licensing terms.
 */

#ifdef HAVE_CONFIG_H
# include <config.h>
#endif

#if STDC_HEADERS
# include <string.h>
# include <stdlib.h>
#else
# if HAVE_STRING_H
#  include <string.h>
# else
#  include <strings.h>
# endif
# if HAVE_STDLIB_H
#  include <stdlib.h>
# else
#  include <malloc.h>
# endif
#endif
#if HAVE_UNISTD_H
# include <sys/types.h>
# include <unistd.h>
#endif
#if HAVE_LIMITS_H
# include <limits.h>
#endif
#if TIME_WITH_SYS_TIME
# include <sys/time.h>
# include <time.h>
#else
# if HAVE_SYS_TIME_H
#  include <sys/time.h>
# else
#  include <time.h>
# endif
#endif

#include <stdarg.h>
#include <stdio.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <ctype.h>
#include <errno.h>
#include "sunos-protos.h"

#include "channels.h"
#include "connections.h"
#include "nicks.h"
#include "config-file.h"
#include "format.h"
#include "dyn-str.h"

/* Pretty-format a string.  Options in the format string can include:
   $a  - the user's IP address
   $A  - the user's hostname, if defined, else IP address.
   $c  - the user's primary channel
   $C  - a list of all the channels the user's currently on
   $d  - the current date and time
   $D  - the date and time the user logged on
   $i  - the time the user's been idle
   $I  - the time the user's primary channel has been idle
   $m  - the user's current client mode
   $n  - the user's nickname
   $p  - the user's current privilege level
   $'? - the contents of "spec_string"s, strings specific to this
         invocation of the command.  $'1 is string 1, $'2 is 2, etc.
   $t  - the topic of the user's primary channel
   $u  - the number of users on the primary channel
   $w  - the number of columns the user is wrapping to
   $$  - a single $ sign.
   (more to come...)

   A specifier can also be preceded with a number, which means that
   the string should be right-justified inside a field of that width,
   if its width is less than that value.  The flag - means left-justify
   instead.

   A specifier can also be preceded with ^, which means to upper-case
   the specifier in question.

   The pointer returned points to static storage, which is not
   preserved between invocations.  If you need to keep it, strdup()
   it.

   (Actually, it's preserved for exactly *one* invocation.  It's kinda
   dumb, but it makes some other stuff a lot cleaner.)

   Returns NULL for fatal errors, e.g. if the string could not be
   malloc'd.

*/

/* A few shortcuts. */
#define fc (clichrs[0])
#define ec (clichrs[1])
#define eec (clichrs[2])
#define fec (clichrs[3])

/* Munge a string so that format characters are escaped properly. */
static char *format_escape(char *s)
{
  char *os, *osp, *clichrs;

  clichrs = get_conf_client_chars();

  os = strdup(s);
  if(!os) return(NULL);
  osp = os;
  while(*osp) {
    if(*osp == ec)      { *(s++) = ec; *(s++) = '1'; }
    else if(*osp == fc) { *(s++) = ec; *(s++) = '2'; }
    else *(s++) = *osp;
    osp++;
  }
  *s = '\0';
  free(os);
  return(s);
}

/* Print an idle time in buffer.  It'll be five characters long, so allocate
   at least 6 chars in the buffer. */
static char *format_idle(char *buffer, time_t delta)
{
  int minutes, hours, days;

  minutes = (delta % 3600) / 60;
  hours = delta / 3600;
  
  if (hours >= 24) {
    days = hours / 24;
    hours %= 24;

    if (days >= 10) {
      sprintf(buffer, "%4dd", days);
    } else {
      sprintf(buffer, "%1dd%2dh", days, hours);
    }
  } else {
    sprintf(buffer, "%d:%02d", hours, minutes);
  }
  return buffer;
}

/* Modifiers and other definitions for format */
#define FORMAT_UPCASE 0x01
#define FORMAT_ALIGN_LEFT (FORMAT_UPCASE << 1)
#define BIGGER_BUF_SIZE 8192 /* Should be sufficient to hold anything,
                                including a full user message */

static char *formata(char *fmt, struct connection *user,
	      int numeric, int num_specs, char **spec_string)
{
  static dyn_string buffer[2] = {DYN_STR_INIT(128), DYN_STR_INIT(128)};
  static char bigger_buf[BIGGER_BUF_SIZE], big2_buf[BIGGER_BUF_SIZE];
  static int n = 1; /* Which buffer we're using this time.  This is so we 
                       can make two consecutive calls to format.  It's a 
                       kludge... */

  dyn_string *p;
  char c, docopy = '\0', minibuf[2], minibuf2[3],
       *clichrs, *copy_to_buf = NULL, *width_field_buf = NULL;
  int i, flag, fmt_pos, modifiers = 0, width = 0;
  time_t curtime;

  n = 1 - n;

  clichrs = get_conf_client_chars();

  p = dyn_strcpy(buffer + n, ""); /* Reset the buffer. */
  if (p == NULL) return NULL;

  if (fmt == NULL) fmt = "[***missing***] $s\n";

  minibuf[1] = '\0'; minibuf2[2] = '\0';

  if(numeric) {
    sprintf(bigger_buf, "%d ", numeric);

    copy_to_buf = bigger_buf;
    p = dyn_strcat(buffer + n, copy_to_buf);
            
    if (p == NULL) return NULL;
  }
  
  for(fmt_pos = 0; fmt[fmt_pos] != '\0'; fmt_pos++) {
    width = 0;
    modifiers = 0;
    docopy = '\0';
    switch (fmt[fmt_pos]) {
    case '$':
      fmt_pos++;
    reflag: /* For when we have a flag argument, and we need to
	       re-switch to find the actual flag. */
      switch(fmt[fmt_pos]) {
      case '\0':
	fmt_pos--;
	minibuf[0] = '$';
	copy_to_buf = minibuf;
	break;
      case '1': case '2': case '3': case '4':
      case '5': case '6': case '7': case '8': case '9':
	/* Way of handling this inspired by NetBSD printf */
	do {
	  width = 10 * width + (fmt[fmt_pos++] - '0');
	} while (isdigit(fmt[fmt_pos]));
	goto reflag;
	break; /* Not reached */
      case '^':
	modifiers |= FORMAT_UPCASE;
	fmt_pos++;
	goto reflag;
	break; /* Not reached */
      case '-':
	modifiers |= FORMAT_ALIGN_LEFT;
	fmt_pos++;
	goto reflag;
	break; /* Not reached */
      case 'a':
	strcpy(bigger_buf, inet_ntoa(user->connection.sin_addr));
	docopy = 'a';
	break;
      case 'A':
	if (user->hostname_resolved > 0)
	  copy_to_buf = user->hostname;
	else
	  copy_to_buf = inet_ntoa(user->connection.sin_addr);
	strcpy(bigger_buf, copy_to_buf);
	docopy = 'A';
	break;
      case 'c':
	sprintf(bigger_buf, "%d", user->channels.primary_channel);
	docopy = 'c';
	break;
      case 'C':
	copy_to_buf = bigger_buf;
	flag = 0;
	*copy_to_buf = '\0';
	for (i = 0; i < num_channels(); i++) {
	  if (user->channels.channel_map & (1 << i)) {
	    sprintf(copy_to_buf, "%s%d", flag?" ":"", i);
	    flag = 1;
	    copy_to_buf = strchr(copy_to_buf, '\0');
	  }
        }
	docopy = 'C';
	break;
      case 'd':
	curtime = time(NULL);
	strftime(bigger_buf, BIGGER_BUF_SIZE, lookup_conf("date_format"),
		 localtime(&curtime));
	docopy = 'd';
        break; 
      case 'D':
	strftime(bigger_buf, BIGGER_BUF_SIZE, lookup_conf("date_format"),
		 localtime(&(user->signon)));
	docopy = 'D';
        break; 
      case 'n':
	strcpy(bigger_buf, nickname(user));
	docopy = 'n';
        break; 
      case 'i':
	format_idle(bigger_buf, time(NULL) - user->last_nonidle);
	docopy = 'i';
        break; 
      case 'I':
	format_idle(bigger_buf, time(NULL) -
		    channels[user->channels.primary_channel].last_nonidle);
	docopy = 'I';
        break; 
      case 'p':
	sprintf(bigger_buf, "%d", user->privs);
	docopy = 'p';
  	break;
      case '\'':
	fmt_pos++;
	i = (int) (fmt[fmt_pos] - '1');
	if((i >= 0) && (num_specs > i)) strcpy(bigger_buf, spec_string[i]);
	  else bigger_buf[0] = '\0';
	docopy = i + '1';
	break;
      case 't':
	strcpy(bigger_buf, 
               get_channel_topic(user->channels.primary_channel) ?
               get_channel_topic(user->channels.primary_channel) : "");
	docopy = 't';
	break;
      case 'u':
	sprintf(bigger_buf, "%d",
                channels[user->channels.primary_channel].connection_count);
	docopy = 'u';
	break;
      case 'w':
	sprintf(bigger_buf, "%d", user->columns); 
	docopy = 'w';
	break;
      case 'm':
        sprintf(bigger_buf, "%08X", user->client_flags);
	docopy = 'm';
        break; 
      default:
        if(fmt[fmt_pos] == fc) {
          minibuf2[0] = ec; minibuf2[1] = fec; copy_to_buf = minibuf2;
        } else if(fmt[fmt_pos] == ec) {
          minibuf2[0] = ec; minibuf2[1] = eec; copy_to_buf = minibuf2;
        } else {
          minibuf[0] = fmt[fmt_pos]; copy_to_buf = minibuf;
        }
	break;
      }
      if(docopy) {
	if(!numeric) copy_to_buf = bigger_buf; else { 
          if(!format_escape(bigger_buf)) return(NULL);
          sprintf(big2_buf, "%c%c%s%cE", fc, docopy, bigger_buf, fc);
          copy_to_buf = big2_buf;
        }
      }
      break;
    case '\n':
      if(numeric && fmt[fmt_pos+1]) {
        sprintf(bigger_buf, "\n%d ", numeric);
        copy_to_buf = bigger_buf;
      } else {
        minibuf[0] = fmt[fmt_pos];
        copy_to_buf = minibuf;
      }
      break;
    default:
      if(fmt[fmt_pos] == fc) {
        minibuf2[0] = ec; minibuf2[1] = fec; copy_to_buf = minibuf2;
      } else if(fmt[fmt_pos] == ec) {
        minibuf2[0] = ec; minibuf2[1] = eec; copy_to_buf = minibuf2;
      } else {
        minibuf[0] = fmt[fmt_pos]; copy_to_buf = minibuf;
      }
      break;
    }

    if (width != 0) {
      if(numeric) width += 4;
      if (modifiers & FORMAT_ALIGN_LEFT) {
	int i = 0;
	
	width_field_buf = malloc(width + 1);
	if (width_field_buf != NULL)
	  for (i = 0; i < width && copy_to_buf[i] != '\0'; i++)
	    width_field_buf[i] = copy_to_buf[i];
	if (i < width) {
	  for (; i < width; i++)
	    width_field_buf[i] = ' ';
	  width_field_buf[i] = '\0';
	  copy_to_buf = width_field_buf;
	} else {
	  free(width_field_buf);
	  width_field_buf = NULL;
	  width = 0;
	}
      } else { /* ! FORMAT_ALIGN_LEFT */
	int i, len;
	for (i = 0; i < width && copy_to_buf[i] != '\0'; i++)
	  ; /* do nothing */
	if (i < width) {
	  len = i;
	  width_field_buf = malloc(width + 1);
	  for (i = 0; i < width - len; i++)
	    width_field_buf[i] = ' ';
	  for (; i < width && copy_to_buf[i - (width - len)] != '\0'; i++)
	    width_field_buf[i] = copy_to_buf[i - (width - len)];
	  width_field_buf[i] = '\0';
	  copy_to_buf = width_field_buf;
	} else {
	  width = 0;
	}
      }
    }

    c = copy_to_buf[0];

    if (modifiers & FORMAT_UPCASE)
      copy_to_buf[0] = toupper(copy_to_buf[0]);

    p = dyn_strcat(buffer + n, copy_to_buf);

    if (modifiers & FORMAT_UPCASE)
      copy_to_buf[0] = c;

    if (p == NULL) return NULL;
  }

  if (width_field_buf) {
    free(width_field_buf);
    width_field_buf = NULL;
    width = 0;
  }
  return buffer[n].buf;
}

char *formatvn(char *conf_name, struct connection *user, int num, ...)
{
  char *s;
  int n, i;
  va_list args;
  char **spec_string = NULL;
  char *blank = "";
  char *r;

  if(num) {
    spec_string = malloc(sizeof(char *) * num);
    if(!spec_string) return NULL;
    va_start(args, num);
    for(i = 0; i < num; i++) {
      spec_string[i] = va_arg(args, char *);
      if(spec_string[i] == NULL) spec_string[i] = blank;
    }
    va_end(args);
  }

  s = lookup_conf_with_numeric(conf_name, &n);
  
  r = formata(s, user, n, num, spec_string);
  if(spec_string) free(spec_string);
  return r;
}

char *formatv(char *fmt, struct connection *user, int n, int num, ...)
{
  int i;
  va_list args;
  char **spec_string = NULL;
  char *blank = "";
  char *r;

  if(num) {
    spec_string = malloc(sizeof(char *) * num);
    if(!spec_string) return NULL;
    va_start(args, num);
    for(i = 0; i < num; i++) {
      spec_string[i] = va_arg(args, char *);
      if(spec_string[i] == NULL) spec_string[i] = blank;
    }
    va_end(args);
  }

  r = formata(fmt, user, n, num, spec_string);
  if(spec_string) free(spec_string);
  return r;
}

/* Strip field codes and numeric from a message. */
char *strip_for_user(char *s)
{
  char *s3 = s, *s2 = s, *clichrs;

  clichrs = get_conf_client_chars();

  s2 = strchr(s2, ' ');
  if(s2) s2++;
  while(s2 && *s2) {
    while(*s2 && *s2 != fc && *s2 != ec && *s2 != '\n') { *(s++) = *(s2++); }
    if(*s2 == ec) {
      s2++;
      *(s++) = (*s2 == eec) ? ec : (*s2 == fec) ? fc : '!';
      s2++;
    } else if(*s2 == fc) { /* *s2 == fc */
      s2 += 2;
    } else if(*s2 == '\n') {
      *(s++) = '\n';
      s2 = strchr(s2, ' ');
      if(s2) s2++;
    }
  }
  *s = '\0';
  return(s3);
}
