/*
 * Initial setup of the chatserver.
 *
 * This module handles (or rather dispatches, mostly) the initial setup of the
 * chatserver.
 *
 * Part of Chatserv, by Jonathan Lennox.  Written for no particularly good
 * reason, because I felt like it.  Allows multiple people to chat over the
 * net in a relatively straightforward environment.  Disclaimer: I figured out
 * how to do TCP/IP programming by reading man pages.
 *
 * Copyright 1995 Jonathan Lennox
 *
 * See LICENSE for licensing terms.
 */

#ifdef HAVE_CONFIG_H
# include <config.h>
#endif

#if STDC_HEADERS
# include <string.h>
#else
# if HAVE_STRING_H
#  include <string.h>
# else
#  include <strings.h>
# endif
#endif
#if TIME_WITH_SYS_TIME
# include <sys/time.h>
# include <time.h>
#else
# if HAVE_SYS_TIME_H
#  include <sys/time.h>
# else
#  include <time.h>
# endif
#endif

/* How many of the rest of these should be autoconf'd? */
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <sys/resource.h>
#include <signal.h>
#ifndef SA_RESTART
/* SunOS doesn't seem to have this...it has SA_INTERRUPT, the complement,
   instead.  We just won't set SA_INTERRUPT, and we should be happy.*/
#define SA_RESTART 0
#endif
#include "sunos-protos.h"

#include "channels.h"
#include "connections.h"
#include "config-file.h"
#include "format.h"
#include "lastlog.h"

/* Initialize the chatserver.  Succeed or exit().
   Return the listening port for incoming connections.
*/

extern char *progname;

void initialize_server(int *incoming_fd, int *ssl_incoming_fd, int *dns_fd)
{
  struct sigaction action;
  char *status;
  int err, retval;
#ifndef __EMX__
#ifndef __CYGWIN32__
  struct rlimit max_descriptors;
#endif
#endif

  /* We want to do this as early as possible so we don't have to free all
     sorts of stuff in the child process. */
  status = "setting up asynchronous DNS server";
  *dns_fd = start_dns();
  if (*dns_fd < 0) { err = *dns_fd; goto abort; }
  
  status = "reading config file";
  err = load_config();
  if (err < 0) { goto abort; }
  else if (err) {
    /* Don't handle this by the abort: bailout -- it's a special case. */
    fprintf(stderr,
	    "%s: syntactic error in config file, line %d.  Aborting.\n",
	    progname, err);
    exit(1);
  }

  status = "setting up channels";
  err = set_up_channels();
  if (err) { goto abort; }

  /* Increase our # of file descriptors to the maximum possible --
     1024 under Solaris.  I believe the number of simultaneous
     connections we can have is this number - 4 (allowing descriptors
     for stdin, stdout, and stderr, and the accepting socket). */

  /* EMX doesn't handle rlimits.  TODO: Come up with some other way
     to limit connections?  Not a worry for a while. */

  /* Neither does CYGWIN. */

#ifndef __EMX__
#ifndef __CYGWIN32__

  status = "getting descriptor count";
  retval = getrlimit(RLIMIT_NOFILE, &max_descriptors);
  if (retval) { err = -errno; goto abort; }

  status = "setting descriptor count";
  max_descriptors.rlim_cur = max_descriptors.rlim_max;
  retval = setrlimit(RLIMIT_NOFILE, &max_descriptors);
  if (retval) { err = -errno; goto abort; }

#endif /* __CYGWIN32__ */
#endif /* __EMX__ */

  /* write(2) can send SIGPIPE if we write to a dead pipe.
     Unfortunately, the default action on a SIGPIPE is to die.  This
     is not what we want, clearly.  Since write(2) also sets errno to
     EPIPE when this happens, we should be okay ignoring SIGPIPE. */
  status = "setting up SIGPIPE handler";
  action.sa_handler = SIG_IGN;
  sigemptyset(&action.sa_mask);
  action.sa_flags = SA_RESTART;
  retval = sigaction(SIGPIPE, &action, NULL);
  if (retval) { err = -errno; goto abort; }

  /* If we catch SIGTERM, we need to send SIGTERM to our child too. */
  status = "setting up SIGTERM handler";
  action.sa_handler = handler;
  /* It'd be nice to do SA_RESETHAND here, but apparently it's not Posix. */
  action.sa_flags = 0;
  retval = sigaction(SIGTERM, &action, NULL);
  if (retval) { err = -errno; goto abort; }

  /* Seed the random number generator. */
  srand((int)time((time_t *)NULL));

  /* Set up the listening port. */
  status = "setting up listening ports";
  *incoming_fd = tcp_listen_on(atoi(lookup_conf("port")));
  *ssl_incoming_fd = tcp_listen_on(atoi(lookup_conf("ssl_port")));
  if (*incoming_fd < 0) { err = *incoming_fd; goto abort; }
  if (*ssl_incoming_fd < 0) { err = *ssl_incoming_fd; goto abort; }

  /* Let people know when the server started. */
  status = "initializing lastlog";
  err = insert_to_log(formatn("server_start_ll_entry",
					   NULL, NULL));
  if (err) { goto abort; }

  /* We're up!  Print this to stderr. */
  fprintf(stderr, "%s: Server initialized, now accepting connections.\n",
	  progname);

  return;

  /* If something screwed up: */
abort:
  if (err < 0)
    fprintf(stderr, "%s: error while %s: %s.  Aborting.\n", 
	    progname, status, strerror(-err));
  else
    fprintf(stderr, "%s: error %d while %s.  Aborting.\n",
	    progname, err, status);

  exit(1);
}
