/*
 * Functions having to do with nicknames.
 *
 * Part of Chatserv, by Jonathan Lennox.  Written for no particularly good
 * reason, because I felt like it.  Allows multiple people to chat over the
 * net in a relatively straightforward environment.  Disclaimer: I figured out
 * how to do TCP/IP programming by reading man pages.
 *
 * Copyright 1995 Jonathan Lennox
 *
 * See LICENSE for licensing terms.
 */

/* AIX requires this to be the first thing in the file... */
#if !defined (__GNUC__) && defined (_AIX)
 #pragma alloca
#endif

#ifdef HAVE_CONFIG_H
# include <config.h>
#endif

#if STDC_HEADERS
# include <string.h>
# include <stdlib.h>
#else
# if HAVE_STRING_H
#  include <string.h>
# else
#  include <strings.h>
# endif
# if HAVE_STDLIB_H
#  include <stdlib.h>
# else
#  include <malloc.h>
# endif
#endif
#ifdef __GNUC__
# undef alloca
# define alloca(size) __builtin_alloca(size)
#else
# if HAVE_ALLOCA_H
#  include <alloca.h>
# else
#  if !defined(_AIX) && !defined(alloca) /* predefined by HP cc +Olibcalls */
char *alloca ();
#  endif
# endif
#endif

#include <stdio.h>
#include <ctype.h>
#include <errno.h>
#include <regex.h>
#include "sunos-protos.h"

#include "channels.h"
#include "connections.h"
#include "nicks.h"
#include "config-file.h"
#include "format.h"
#include "lastlog.h"

/* Look up (case-insensitive) the given nickname.  NULL if it doesn't
   exist. */
struct connection *lookup_nickname(char *name)
{
  struct connection *user;

  for (user = connections; user != NULL; user = user->next) {
    if (strcasecmp(name, user->nick) == 0)
      break;
  }

  return user;
}

char *nickname(struct connection *user)
{
  return user->nick;
}

/* Change the user's nickname to the given string.
   Return value:
     0 if okay
     1 if no change to the nick
     2 if nick is in use by someone else
     3 if nick is too long
     4 if nick is invalid for some other reason
     -1 on other error
*/
int change_nickname(struct connection *user, char *new_nick)
{
  struct connection *current_holder;
  char *tmp_string;
  int i;

  /* Nicks must be nonempty, */
  if (new_nick[0] == '\0') return 4;

  /* be made up of visible characters, */
  for (i = 0; new_nick[i] != '\0'; i++)
    if (!isgraph(new_nick[i])) return 4;
    
  /* not be too long, */
  if (i >= MAX_NICK_LENGTH) 
    return 3;
    
  /* -mbg- not match the bad nick regexp, */
  if(regexec(get_conf_regex(BADNICK_REGEX), new_nick, 0, NULL, 0) == 0)
    return 4;

  /* and not be the default nick. */

  current_holder = lookup_nickname(new_nick);
  if (current_holder == NULL || current_holder == user) {
    if (strcmp(new_nick, user->nick) == 0) 
      /* They tried to change their nick to what it already is. */
      return 1;
    tmp_string = (char *) malloc(strlen(new_nick) + 1);
    if (tmp_string == NULL) return -1;
    strcpy(tmp_string, new_nick);
    free(user->nick);
    user->nick = tmp_string;
    return 0;
  }
  else return 2;
}
    
int nickname_cmd(char *cmd_name, struct connection *user, char *args)
{
  char *old_nick = alloca(strlen(nickname(user)) + 1);
  int retval;

  if (old_nick == NULL) return -ENOMEM;
  strcpy(old_nick, nickname(user));

  if (strcmp(args, "") == 0) {
    return write_to_user(formatn("current_nick", user, NULL),
			 user);
  }

  retval = change_nickname(user, args);
  if (retval == 1) {
    /* Nick was not changed. */
    return write_to_user(formatn("nick_no_change", user, NULL),
			 user);
  } else if (retval == 2) {
    /* Nick is in use. */
    return write_to_user(formatn("nick_in_use", user, args),
			 user);
  } else if (retval == 3) {
    /* Nick is too long. */
    return write_to_user(formatn("nick_too_long", user, args),
			 user);
  } else if (retval) {
    /* Nick was invalid for some other reason */
    return write_to_user(formatn("bad_nick", user, args),
			 user);
  } else {
    if (user->status == normal) {
      retval = visible_action(user);
      if (retval) return retval;
    
      retval = insert_to_log(formatn("nickchange_ll_entry",
				    user, old_nick));
      if (retval) return retval;
    }
    if (user->status == unnamed) {
      retval = write_to_user(formatn("first_nick", user, args),
			     user);
      if (retval) return retval;
      return greet_new_user(user);
    } else if (user->channels.channel_map == 0) {
      /* If the user isn't on any channels, just write to the user. */
      return write_to_user(formatn("nick_changed",
				  user, old_nick),
			   user);
    } else {
      return write_to_channelmap(formatn("nick_changed",
					user, old_nick),
				 user->channels.channel_map);
    }
  }
  return 0;
}

#if DO_UNORTHODOX

#define MAX_UNORTHODOX 16

/* Set someone Unorthodox. */
/* It's generalizable now.... oh yes.... */
int trudy_cmd(char *cmd_name, struct connection *user, char *args)
{
  char *rest, *alias, **p, *s, *aliasend, quotechar;

  rest = args;

    while(*rest && (user->unorthocount < MAX_UNORTHODOX)) {
      alias = rest;
      quotechar = *alias;
      if(isalnum(quotechar)) {
        aliasend = strchr(alias, ' ');
      } else {
        alias++;
        aliasend = strchr(alias, quotechar);
      }
      if(!aliasend) aliasend = strchr(alias, '\0');
      if(aliasend) {
        *aliasend = '\0';
        user->unorthocount++;
        p = realloc(user->unorthoname,
                    sizeof(void *) * user->unorthocount);
        if(!p) {
          user->unorthocount--;
          return -ENOMEM;
        }
        user->unorthoname = p;
        s = malloc(MAX_NICK_LENGTH);
        if(!s) {
          user->unorthoname[user->unorthocount - 1] = NULL;
          return -ENOMEM;
        }
        user->unorthoname[user->unorthocount - 1] = s;
        strncpy(user->unorthoname[user->unorthocount - 1],
                alias, MAX_NICK_LENGTH);
        rest = aliasend + 1;
        while(isspace(*rest)) rest++;
      }
    }
    if(user->unorthocount) {
      user->unorthodox = 1;
      return write_to_user(formatn("unorthodox_on",
  				    user, NULL),
  			   user);
    }
  return 0;
}
    
int notrudy_cmd(char *cmd_name, struct connection *user, char *args)
{
  int i;
  char *rest;

  rest = args;

    if(user->unorthodox) {
      user->unorthodox = 0;
      if(user->unorthocount) {
        for(i = 0; i < user->unorthocount; i++)
          free(user->unorthoname[i]);
        free(user->unorthoname);
      }
      user->unorthocount = 0;
      user->unorthoname = NULL;
      return write_to_user(formatn("unorthodox_off",
				    user, NULL),
			   user);
    }
  return 0;
}
#endif
