/*
 * TLS connection handling.
 *
 * Part of Chatserv, by Jonathan Lennox.  Written for no particularly good
 * reason, because I felt like it.  Allows multiple people to chat over the
 * net in a relatively straightforward environment.  Disclaimer: I figured out
 * how to do TCP/IP programming by reading man pages.
 *
 * Copyright 2005 Matthew Gerber
 *
 * See LICENSE for licensing terms.
 */

/* AIX requires this to be the first thing in the file... */
#if !defined (__GNUC__) && defined (_AIX)
 #pragma alloca
#endif

#ifdef HAVE_CONFIG_H
# include <config.h>
#endif

#if STDC_HEADERS
# include <string.h>
# include <stdlib.h>
#else
# if HAVE_STRING_H
#  include <string.h>
# else
#  include <strings.h>
# endif
# if HAVE_STDLIB_H
#  include <stdlib.h>
# else
#  include <malloc.h>
# endif
#endif
#if HAVE_UNISTD_H
# include <sys/types.h>
# include <unistd.h>
#endif
#if TIME_WITH_SYS_TIME
# include <sys/time.h>
# include <time.h>
#else
# if HAVE_SYS_TIME_H
#  include <sys/time.h>
# else
#  include <time.h>
# endif
#endif
#if HAVE_LIMITS_H
# include <limits.h>
#endif
#ifdef __GNUC__
# undef alloca
# define alloca(size) __builtin_alloca(size)
#else
# if HAVE_ALLOCA_H
#  include <alloca.h>
# else
#  if !defined(_AIX) && !defined(alloca) /* predefined by HP cc +Olibcalls */
char *alloca ();
#  endif
# endif
#endif

#include <stdio.h>
#include "channels.h"
#include "connections.h"
#include "config-file.h"

#ifdef GNUTLS11
static gnutls_anon_server_credentials credentials;
static gnutls_certificate_credentials xcred;

static gnutls_dh_params dh_params;

static void ssl_generate_dh_params(void) {
  gnutls_dh_params_init( &dh_params);
  gnutls_dh_params_generate2( dh_params, 1024);   
}

void ssl_init() {
  char *certfile = lookup_conf("ssl_cert");
  char *keyfile = lookup_conf("ssl_key");

  if(gnutls_global_init()) {
    fprintf(stderr, "Error initializing gnutls.\n");
    exit(1);
  }

  ssl_generate_dh_params();
  if(gnutls_anon_allocate_server_credentials(&credentials)) { 
    fprintf(stderr, "Could not allocate server credentials.\n");
    exit(1);
  }
  gnutls_anon_set_server_dh_params(credentials, dh_params);

  if(gnutls_certificate_allocate_credentials(&xcred)) {
    fprintf(stderr, "Could not allocate server credentials.\n");
    exit(1);
  }
  gnutls_certificate_set_x509_key_file(xcred, certfile, keyfile, 
				       GNUTLS_X509_FMT_PEM);
  gnutls_certificate_set_dh_params(xcred, dh_params);
  printf("Secure Sockets Layer initialization complete.\n");
}

void ssl_end(SSL ssl) {
  gnutls_deinit(*ssl);
}

int ssl_write(SSL ssl, char *s, int len) {
  return gnutls_write(*ssl, s, len);
}

int ssl_read(SSL ssl, char *s, int len) {
  return gnutls_read(*ssl, s, len);
}

struct connection *start_ssl_setup(struct connection *newcon, int fd) {
  gnutls_session *session;
  int kx_priority[] = { GNUTLS_KX_DHE_RSA, GNUTLS_KX_RSA, 0 };

  session = malloc(sizeof(gnutls_session));

  if(!session) {
    close_connection(newcon);
    return NULL;
  }

  if (gnutls_init(session, GNUTLS_SERVER)) {
    close_connection(newcon);
    return NULL;
  }

  gnutls_set_default_priority(*session);
  gnutls_kx_set_priority(*session, kx_priority);
  gnutls_credentials_set(*session, GNUTLS_CRD_CERTIFICATE, xcred);
  gnutls_credentials_set(*session, GNUTLS_CRD_ANON, credentials);
  gnutls_dh_set_prime_bits(*session, 1024);

  gnutls_transport_set_ptr(*session, (gnutls_transport_ptr) fd);

  newcon->ssl_finished = 0;
  newcon->ssl = session;
  continue_ssl_setup(newcon);
  return newcon;
}

void continue_ssl_setup(struct connection *current) {
  int retval;

  retval = gnutls_handshake(*(current->ssl));
  if(retval == 0) {
    current->ssl_finished = 1;
  } else {
    if(retval == GNUTLS_E_AGAIN || retval == GNUTLS_E_INTERRUPTED) {
      current->ssl_finished = 0;
    } else {
      fprintf(stderr, "SSL FATAL: ");
      gnutls_perror(retval);
      close_connection(current);
    }
  }
}
#endif

#ifndef DO_SSL
void ssl_init() {}
void ssl_end(SSL ssl) {}
int ssl_write(SSL ssl, char *s, int len) {}
int ssl_read(SSL ssl, char *s, int len) {}
struct connection *start_ssl_setup(struct connection *newcon, int fd) {}
void continue_ssl_setup(struct connection *current) {}
#endif
